/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 920 $ $LastChangedDate: 2002-09-10 10:03:29 +0200 (Tue, 10 Sep 2002) $
 *
 * Chini - check if initialized.
 */

open Ast;
open Util;
open Tyutil;

/* 
   This structure is returned by all chini_* functions. It contains list
   of initialized symbols along with information under what circumstances
   they might be considered initialized. 

   ,,Circumstances'' above are related to loops. This is because symbol
   might be initialized after conditional `break FOO' statement, and should not
   be considered initialized when we finish processing FOO loop. Therefore
   loops are assigned numbers, and so symbols returned in this structure are.

   try { } with { } blocks are treated similarly, as they can be exited
   with raise statement.
   
   Example:

    foo: while (...) {
	x = 1;
	bar: while (true) {
	    y = 1;
	    if (...)
		break bar;
	    z = 1;
	    if (...)
		break foo;
	    u = 1;
	}
	v = 1;
    }

   Loop foo has number 1, loop bar has number 2. TODO: finish it.
   
 */

struct ini_lst {
    // if syms(s) == N, then s will be initialized before leaving loop 
    // level N+1. Function has loop_level == 0
    <var_symbol, int>Ptrhash::t syms;

    // Lowest level we *will* jump out after executing stmts in
    // question. 0 for function-level (i.e. return stmt),
    // 1 for 1st level loop and so on. -1 if we don't know for sure
    // if we'll jump anywhere.
    int will_break_at;

    // can_break_at(N) is true iff stmt can break at given loop level
    Bitvector::t can_break_at;
}

struct chini_ctx {
    // max(try_level, loop_level)
    int level;
    int try_level;
    int loop_level;

    // maximal loop level we can terminate at
    int can_break_at;

    bool all_globals_initialized;

    // initialized symbols, along with loop_levels
    <var_symbol, int>Ptrhash::t initialized;

    <expr>option assign_target_expr;
}

<stmt>list inits;
<var_symbol, void>Ptrhash::t globals;
bool chini_globals;
bool globally_safe;

section init 
{
    inits = Nil;
    globals = Ptrhash::create(13);
    chini_globals = false;
    globally_safe = false;
}

chini_ctx copy_context(chini_ctx c)
{
    return c with {initialized = Ptrhash::copy(c.initialized)};
}

ini_lst nothing_initialized()
{
    return {will_break_at = -1, 
	    syms = Ptrhash::create(5),
	    can_break_at = Bitvector::make(1, false)};
}

void update_break_ctx(chini_ctx c, ini_lst lst)
{
    int i;

    for (i = c.can_break_at + 1; i < Bitvector::length(lst.can_break_at); i++)
	if (Bitvector::get(lst.can_break_at, i))
	    c.can_break_at = i;
}

void update_ctx(chini_ctx c, ini_lst lst)
{
    void add(var_symbol s, int lev)
    {
//dbg("uctx: " + symbol_name(s) + " @ " + itoa(Hacks::ptr_of(c.initialized)) + "\n");
	if (Ptrhash::mem(c.initialized, s)) {
	    let oldlev = Ptrhash::find(c.initialized, s) in
	    if (oldlev > lev) {
		Ptrhash::remove(c.initialized, s);
		Ptrhash::add(c.initialized, s, lev);
	    }
	} else {
	    Ptrhash::add(c.initialized, s, lev);
	}
    }
    
    Ptrhash::iter(lst.syms, add);
    
    update_break_ctx(c, lst);
}

// Marge [xnew] into [base], return updated [base] (in-place).
// In [branch] mode it roughly does [(base & xnew)], in non-branch 
// [(base | xnew)].
//
// Non-branch mode: if variable is in both [base] and [xnew], but in [xnew] it
// is at the lower level, it's moved to that level in [base]. If variable is
// only in [base] it is left were it was. If variable is only in [xnew], it's
// added to [base] at the same level.
//
// Branch-mode: If variable is only in one of [xnew] and [base], it's 
// discarded. Otherwise it's placed at upper level of the two. However
// if [xnew.will_break_at != -1] we do not touch [base]'s
// levels >= [xnew.will_break_at].

ini_lst merge(ini_lst base, ini_lst xnew)
{
    <var_symbol, int>Ptrhash::t ret = 
	Ptrhash::create(Ptrhash::length(base.syms) + Ptrhash::length(xnew.syms));

    void add(var_symbol s, int lev)
    {
	try {
	    int new_lev = Ptrhash::find(xnew.syms, s);
	    if (new_lev < lev)
		lev = new_lev;
	    Ptrhash::remove(xnew.syms, s);
	} with {
	case Not_found:
		Ptrhash::add(ret, s, lev);
	}
    }

    Ptrhash::iter(base.syms, add);
    Ptrhash::iter(xnew.syms, fun (s, lev) (Ptrhash::add(ret, s, lev)));

    base.syms = ret;

    Bitvector::or(base.can_break_at, xnew.can_break_at);

    return base;
}

ini_lst merge_branch(ini_lst base, ini_lst xnew)
{
    if (base.will_break_at == 0 && xnew.will_break_at != 0)
	return merge_branch(xnew, base);
	
    if (xnew.will_break_at == -1)
	base.will_break_at = -1;

    if (base.will_break_at != -1 && xnew.will_break_at > base.will_break_at)
	base.will_break_at = xnew.will_break_at;

    <var_symbol, int>Ptrhash::t ret = 
	Ptrhash::create(Ptrhash::length(base.syms) + Ptrhash::length(xnew.syms));

    void add(var_symbol s, int lev)
    {
	if (xnew.will_break_at != -1 && lev >= xnew.will_break_at)
	    Ptrhash::add(ret, s, lev);
	else
	    try {
		let new_lev = Ptrhash::find(xnew.syms, s) in
		if (new_lev > lev)
		    lev = new_lev;
		Ptrhash::add(ret, s, lev);
	    } with { 
	    case Not_found:
		if (xnew.will_break_at != -1)
		    Ptrhash::add(ret, s, xnew.will_break_at);
	    }
    }

    Ptrhash::iter(base.syms, add);

    base.syms = ret;
    
    Bitvector::or(base.can_break_at, xnew.can_break_at);
    
    return base;
}

/*
void dbg_dump_ini_lst(ini_lst l)
{
    print_string("(ini_lst will_break_at = " + itoa(l.will_break_at) + 
	     ", syms = ( ");
    
    void prt_one(var_symbol s, int lev)
    {
	print_string(symbol_name(s) + ":" + itoa(lev) + " ");
    }

    Ptrhash::iter(l.syms, prt_one);

    print_string("))\n");
}
*/

ini_lst chini_seq(chini_ctx c, *(ini_lst (chini_ctx, <'a>located)) chini_one, 
		  <<'a>located>list data)
{
    ini_lst r = nothing_initialized();
    bool did_warn = false;

    void do_one(<'a>located d)
    {
	let rr = chini_one(c, d) in {
//dbg("one: ");
//dbg_dump_ini_lst(r);
	    if (r.will_break_at == -1) {
		r.will_break_at = rr.will_break_at;
	    } else {
		if (flags.w_dead && !did_warn) {
		    warnl(d.loc, "dead code");
		    did_warn = true;
		}
	    }
	    update_ctx(c, rr);
	    r = merge(r, rr);
	}
    }

    List::iter(do_one, data);

//dbg("seq: ");
//dbg_dump_ini_lst(r);

    return r;
}

// We're not at level [n], fixup [l] as aproptiate 
ini_lst cut_down(int n, ini_lst l)
{
    <var_symbol, int>Ptrhash::t ret = 
	Ptrhash::create(Ptrhash::length(l.syms));
    
    void add(var_symbol s, int lev)
    {
	if (lev <= n)
	    Ptrhash::add(ret, s, lev);
    }
    
    Ptrhash::iter(l.syms, add);
    l.syms = ret;
    
    Bitvector::resize(l.can_break_at, n + 1, false);
    if (l.will_break_at == n + 1)
	l.will_break_at = -1;

    return l;
}

ini_lst chini_branches(chini_ctx c, <stmt>list branches)
{
    ini_lst fold(<stmt>list br, ini_lst il)
    {
//dbg("branches_got: ");
//dbg_dump_ini_lst(il);

	switch br {
	case Cons(s, ss):
	    let cc = copy_context(c) in
	    let is = chini_stmt(cc, s) in
	    return fold(ss, merge_branch(il, is));
	    
	case Nil:
	    return il;
	}
    }

    let r = nothing_initialized() in {
	r.will_break_at = 0;
	return fold(branches, r);
    }
}

void check_if_all_globals_initialized(chini_ctx c, string msg)
{
    if (c.all_globals_initialized)
	return;
    
    <var_symbol>list nok = Nil;
    
    _ check(_ s, _)
    {
	if (!Ptrhash::mem(c.initialized, s))
	    nok = Cons(s, nok);
    }
    
    Ptrhash::iter(globals, check);

    if (List::length(nok) != 0) 
	error(msg + "following global symbols might be uninitalized: " + 
	      String::concat(", ", List::map(symbol_name, nok)));

    // always set, to prevent repeating error messages
    c.all_globals_initialized = true;
}

ini_lst chini_symbol(chini_ctx c, var_symbol v)
{
    if (has_property(v, Is_initialized) ||
	has_property(v, Is_union_field) ||
	has_property(v, Is_in_pattern) ||
        (v.a.scope_level <= 0 && v.module != Util::current_module))
	return nothing_initialized();	    // ok

    if (v.a.scope_level <= 0) {
	// We have access to global symbol from current module here.
	// If it is function, that was already checked not to use global
	// variables (i.e. has Is_globally_safe set) or it's Extern_c symbol
	// we don't complain or clear globally_safe flag for current function.
	// extern_c stuff is assumed to know, what it is doing.
	if (has_property(v, Is_globally_safe) || // already checked, and ok
	    has_property(v, Is_extern_c)) 
	    return nothing_initialized();
	else if (chini_globals) {
	    // otherwise if we are checking init section:
	    if (is_global_function(v)) {
		// we allow calling functions from current module only after
		// *all* global symbols has been initialized.
		check_if_all_globals_initialized(c, "cannot call function `" + 
						 symbol_name(v) + "' because ");
	    
		// pretend nothing happened
		return nothing_initialized();
	    }
	} else {
	    // we are not checking init section, so just clear safe-flag.
	    globally_safe = false;
	    return nothing_initialized();
	}
    }
    
    try {
	Ptrhash::find(c.initialized, v);
    } with {
    case Not_found:
	error("variable " + symbol_name(v) + " might be used uninitalized");
    }

    return nothing_initialized();
}

ini_lst chini_fnc(chini_ctx c, lambda_record l, string name)
{
    ini_lst i = chini_stmt(c, l.lambda_body);

    if (!is_void_func_type(l.lambda_type) && i.will_break_at != 0)
	error(name + " can finish without returning a value");
    
    return nothing_initialized();
}

ini_lst chini_assign(chini_ctx c, expr e1, expr e2)
{
    ini_lst i1 = nothing_initialized(), i2;
    <var_symbol>list syms = Nil;
    
    void collect_symbols(expr e)
    {
	switch e.a.raw {
	case Ref(rr):
	    var_symbol s = get_binding(rr.ref_symbol);
	    int lev = c.can_break_at + 1;
	    
	    if (Ptrhash::mem(i1.syms, s))
		return;
	    
	    try {
		lev = Ptrhash::find(c.initialized, s);
	    } with { case Not_found: skip; }
	
	    if (lev > c.can_break_at)
		Ptrhash::add(i1.syms, s, c.can_break_at);
	    
	case Tuple_expr(l):
	    List::iter(collect_symbols, l);

	case _:
	    skip;
	}
    }
    
    collect_symbols(e1);

    // If e2 is lambda expression, we update context first, as there
    // is no way it can be called before this assignment is done. 
    // Otherwise local functions cannot be made self-recursive.
    // Note that it still doesn't allow mutual recursion which,
    // generally speaking, sucks eggs. 
    switch e2.a.raw {
    case Lambda(_):
	update_ctx(c, i1);
	i2 = chini_expr(c, e2);
    case _:
	i2 = chini_expr(c with {assign_target_expr = Some(e1)}, e2);
	update_ctx(c, i1);
    }

    update_ctx(c, i2);

    return merge(i1, i2);
}

ini_lst chini_nl_expr(chini_ctx c, nl_expr ex)
{
    ini_lst seq(<expr>list l)
    {
	return chini_seq(c, chini_expr, l);
    }
    
    switch ex.raw {
    case Assign(e1, e2):
	return chini_assign(c, e1, e2);
    
    case Assign_target:
	switch c.assign_target_expr {
	case Some(e):
	    return chini_expr(c, e);
	case None:
	    return ice("Chini::chini_nl_expr(): no parent assign");
	}
    
    case Simple_binop(_, e1, e2):
	return seq([e1, e2]);
	
    case Simple_unop(_, e):
	return chini_expr(c, e);
    
    case Ref(rr):
	return chini_symbol(c, get_binding(rr.ref_symbol));

    case Call(e, el):
	return seq(Cons(e, el));

    case Tuple_expr(el):
	return seq(el);

    case Union_expr(u):
	return chini_expr(c, u.nx_expr);
    
    case Lambda(l):
	return chini_fnc(c, l, "this lambda-expression");

    case Struct_expr(s, _):
	return seq(List::map(fun expr (sini_expr se) (se.sx_expr), s));

    case Struct_copy(e, s, _):
	return seq(Cons(e, List::map(fun expr (sini_expr se) (se.sx_expr), s)));

    case Dot_expr(de, _):
	return chini_expr(c, de.sx_expr);
    
    // contants arn't really interesting in this context ;)
    case Bool_const(_):
    case Int_const(_):
    case Float_const(_):
    case String_const(_):
    case Void_const:
    case Null_const:	
	return nothing_initialized();
    }
}

ini_lst chini_expr(chini_ctx c, expr ex)
{
    return locate(partial(chini_nl_expr, c), ex);
}

void can_break_at(ini_lst i, int lev)
{
    if (Bitvector::length(i.can_break_at) <= lev)
	Bitvector::resize(i.can_break_at, lev + 1, false);
    
    Bitvector::set(i.can_break_at, lev, true);
}

<*(pattern, bool)>list collect_clauses(<case_clause>list cases)
{
    *(pattern, bool) convert(*(pattern, expr) oc)
    {
	let (pat, cond) = oc in
	switch cond.a.raw {
	case Bool_const(true):
	    return (pat, /* can_fail= */ false);
	case _:
	    return (pat, true);
	}
    }
    
    switch cases {
    case Nil:
	return Nil;
    case Cons(cc, tl):
	return List::concat(List::map(convert, cc.cases), collect_clauses(tl));
    }
}

void check_conds(chini_ctx c, <case_clause>list ccs)
{
    switch ccs {
    case Cons(cc, tl):
	void chini_one_case(*(pattern, expr) oc)
	{
	    let (_, cond) = oc in
	    chini_expr(c, cond);
	}
	
	// someday-TODO: we happily discard whatever condition 
	// might initialize
	List::map(chini_one_case, cc.cases);

	check_conds(c, tl);

    case Nil:
	skip;
    }
}

ini_lst chini_nl_stmt(chini_ctx c, nl_stmt nlr)
{
    stmt get_body(case_clause cc)
    {
	return cc.body;
    }

    switch nlr {
    case Vardef(_):
    case Skip:
	return nothing_initialized();
    
    case Expr(e):
	return chini_expr(c, e);
    
    case Return(e):
	let r = chini_expr(c, e) in {
	    r.will_break_at = 0;
	    return r;
	}

    case Block(lst):
	return chini_seq(c, chini_stmt, lst);

    case Pair(s1, s2):
	return chini_seq(c, chini_stmt, [s1, s2]);
    
    case Switch(e, clauses):
	Patrange::check_exhaustiveness(collect_clauses(clauses));
	
	let exret = chini_expr(c, e) in {
	    update_ctx(c, exret);
	    check_conds(c, clauses);
	    let b = List::map(get_body, clauses) in
	    return merge(chini_branches(c, b), exret);
	}
	
    case Flying_let(assigns):
    case Let(assigns, _):
	void check_assign(*(pattern, expr) x)
	{
	    let (pat, _) = x in
	    let pats = [(pat, false)] in
	    Patrange::check_exhaustiveness(pats);
	}
	
	List::iter(check_assign, assigns);
	
	let a = chini_seq(c, chini_expr, List::map(snd, assigns)) in
	switch nlr {
	case Let(_, s):
	    return merge(chini_stmt(c, s), a);
	case _:
	    return a;
	}
    
    case If(e, s1, s2):
	let exret = chini_expr(c, e) in {
	    update_ctx(c, exret);
	    return merge(chini_branches(c, [s1, s2]), exret);
	}
    
    case Loop(l):
	chini_ctx lc = copy_context(c);
	
	let lev = lc.level + 1 in {
	    lc.level = lev;
	    (get_binding(l.ll_symbol)).a.loop_level = lev;
	    lc.loop_level = lev;
	}
	
	let body_ini = chini_stmt(lc, l.ll_body) in
	let can_terminate = lc.can_break_at >= lc.loop_level in
	let ret = cut_down(c.level, body_ini) in {
	    // it might never be executed, so we ignore the result,
	    // except for the can_break part
	    ini_lst inc = chini_stmt(c, l.ll_inc);
	    Bitvector::or(ret.can_break_at, inc.can_break_at);

	    if (!can_terminate) {
		// If the loop cannot terminate any code hence forth is dead
		// and we *will* break at least at c.can_break_at.
		update_break_ctx(c, ret);
		ret.will_break_at = c.can_break_at;
	    }
	    
	    return ret;
	}
    
    case Break(b):
	let r = nothing_initialized() in {
	    r.will_break_at = (get_binding(b.break_symbol)).a.loop_level;
	    can_break_at(r, r.will_break_at);
	    return r;
	}
	
    case Break_if_not(e):
	let r = chini_expr(c, e) in {
	    switch e.a.raw {
	    case Bool_const(false):
		r.will_break_at = c.loop_level;
		can_break_at(r, c.loop_level);
	    case Bool_const(true):
		skip;	// well, we could not break here
	    case _:
		can_break_at(r, c.loop_level);
	    }

	    return r;
	}
    
    case Open(_):
	return nothing_initialized();
    
    case Try(tr):
	chini_ctx tc = copy_context(c);
	tc.try_level = ++tc.level;

	let body_ret = chini_stmt(tc, tr.try_body) in
	let restore = body_ret.will_break_at == c.level + 1 in
	let cut_ret = cut_down(c.level, body_ret) in
	let with_matching = List::map(get_body, tr.try_with) in
	let with_ret = chini_branches(c, with_matching) in
	let ret = merge_branch(cut_ret, with_ret) in
	let fc = copy_context(c) in {
//dbg("brwba= " +itoa(body_ret.will_break_at)+"\n");
	    if (restore)
		    ret.will_break_at = with_ret.will_break_at;

	    // statements in finally are *always* executed, no
	    // matter how deep loop we happen to jump out
	    fc.can_break_at = c.can_break_at;
	    let fret = chini_stmt(fc, tr.try_finally) in
	    return merge(ret, fret);
	}
    
    case Raise(e):
	let r = chini_expr(c, e) in {
	    r.will_break_at = c.try_level;
	    can_break_at(r, c.try_level);
	    return r;
	}
    }
}

ini_lst chini_stmt(chini_ctx c, stmt s)
{
    return locate(partial(chini_nl_stmt, c), s);
}

chini_ctx make_ctx()
{
    return {level = 0,
	    try_level = 0,
	    loop_level = 0,
	    can_break_at = 0,
	    initialized = Ptrhash::create(13),
	    all_globals_initialized = false,
	    assign_target_expr = None};
}

void chini_nl_decl(nl_decl d)
{
    switch d {
    case Func_decl(fn):
	globally_safe = true;
	chini_fnc(make_ctx(), fn.fn_lambda, 
		  qualified_name(fn.fn_symbol) + "()");
	if (globally_safe)
	    set_property(get_binding(fn.fn_symbol), Is_globally_safe);
    
    case Section_decl(n, l):
	chini_fnc(make_ctx(), l, "this section");
	switch n {
	case "init":
	    inits = Cons(l.lambda_body, inits);
	case _:
	    skip;
	}
    
    case Global_decl(v):
	Ptrhash::add(globals, get_binding(v.var_symbol), ());
	
    case Top_level_open(_):
    case Extern_c(_, _):
    case Func_proto(_):
    case Tydef(_):
    case Exception_decl(_):
	skip;
    }
}

void chini_decl(decl d)
{
    locate(chini_nl_decl, d);
}

void chini_inits()
{
    _ s = Block(List::rev(inits));
    _ c = make_ctx();

    chini_globals = true;
    
    ignore(chini_nl_stmt(c, s));
    
    check_if_all_globals_initialized(c, "problems with init sections: ");
}

// vim:sts=4:sw=4
