/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 880 $ $LastChangedDate: 2002-09-07 15:37:29 +0200 (Sat, 07 Sep 2002) $
 *
 * Module - Module save/restore and symbol lookup. 
 *
 * Modules are serialized to XML. Serialization of ty type is in Xmldump.
 */

open Module;
open Ast;
open Util;
open Fileutil;
open Xmldump;

<string>list search_path = ["."];
<module>Hashtbl::t open_modules = Hashtbl::create(0);
string empty_module = "";

void add_to_search_path(string dir)
{
    search_path = Cons(dir, search_path);
}

void assume_module_is_empty(string name)
{
    empty_module = name;
}

ty_symbol lookup_tysym(id module, id name)
{
    module m = open_module(module);
    return Hashtbl::find(m.mod_tydefs, name);
}

var_symbol lookup_var(id module, id name)
{
    module m = open_module(module);
    return Hashtbl::find(m.mod_vars, name);
}

<sf_symbol>list lookup_struct_field(id module, id name)
{
    module m = open_module(module);
    return Hashtbl::find_all(m.mod_struct_fields, name);
}

ty_symbol lookup_tysym_q(module mod, string name)
{
    switch rev_qualified_name(name) {
    case Simple_name(n):
	return Hashtbl::find(mod.mod_tydefs, n);
	
    case Qualified_name(m, n):
	if (m == mod.mod_name)
	    return Hashtbl::find(mod.mod_tydefs, n);
	else
	    return lookup_tysym(m, n);
	    
    case Symbol(_):
	return ice("Module::lookup_tysym_q(): rev_qualified_name went crazy");
    }
}

void unserialize_all(string name, Xml::node root, module m,
	     *(void (module, Xml::node)) unserialize_one)
{
    void f(Xml::node n)
    {
	unserialize_one(m, n);
    }

    switch Xml::find_element(name, root) {
    case Xml::Element(_, children, _):
	List::iter(f, children);

    case _:
	raise Not_found;
    }
}

void cross_bind(module m, ty t)
{
    void f(ty tt)
    {
	switch tt {
	case Named_type(nt):
	    let tydef = lookup_tysym_q(m, qualified_name(nt.tydef)) in 
	    nt.tydef = Symbol(tydef);

	case _:
	    skip;
	}
    }

    type_iter(f, t);
}

<'a>symbol common_unserialize(module m, Xml::node n, <ty>option maybe_ty, 'a data)
{
    location loc = {file = Xml::prop_value("file", n),
		    first_line = atoi(Xml::prop_value("line", n)),
		    first_column = atoi(Xml::prop_value("column", n)),
		    last_line = 0,
		    last_column = 0};
    ty t;

    switch maybe_ty {
    case Some(tt):
	t = tt;
	
    case None:
	t = unxmlize_ty(Xml::find_element("type", n));
	cross_bind(m, t);
    }
	
    return {name = Xml::prop_value("name", n),
	    stype = t,
	    sloc = loc,
	    module = m.mod_name,
	    a = data};
}

void unserialize_type1(module m, Xml::node n)
{
    let tyvars = unxmlize_tyvars(Xml::find_element("tyvars", n)) in
    let ty_data = {tysym_tyvars = tyvars} in
    let tysym = common_unserialize(m, n, Some(Abstract_type), ty_data) in
    Hashtbl::add(m.mod_tydefs, tysym.name, tysym);
}

void unserialize_type2(module m, Xml::node n)
{
    let tysym = Hashtbl::find(m.mod_tydefs, Xml::prop_value("name", n)) in {
	tysym.stype = unxmlize_ty(Xml::find_element("type", n));
	cross_bind(m, tysym.stype);
    }
}

void unxmlize_props(module m, var_symbol sym, Xml::node nn)
{
    void unxmlize_prop(Xml::node n)
    {
	_ p;
	
        // any change here should be reflected in xmlize_props() below
	switch n {
	case Xml::Element("is-union-field", _, _):
	    p = Is_union_field;

	case Xml::Element("is-lvalue", _, _):
	    p = Is_lvalue;

	case Xml::Element("is-initialized", _, _):
	    p = Is_initialized;

	case Xml::Element("is-global-var", _, _):
	    p = Is_global_var;

	case Xml::Element("is-exception", _, _):
	    p = Is_exception;

	case Xml::Element("is-exported", _, _):
	    p = Is_exported;

	case Xml::Element("is-extern-c", _, _):
	    p = Is_extern_c;

	case Xml::Element("parent-union", _, _):
	    let ts = lookup_tysym_q(m, Xml::prop_value("value", n)) in {
		sym.a.parent_union = Some(ts);
		return;
	    }
	
	case _:
	    raise Invalid_argument("Module::unxmlize_props");
	}

	set_property(sym, p);
    }

    switch nn {
    case Xml::Element("props", children, _):
	List::iter(unxmlize_prop, children);

    case _:
	raise Invalid_argument("Module::unxmlize_props");
    }
}

void reference_self_from_parent_struct(sf_symbol sym, ty_symbol parent)
{
    id sn = simple_name(Symbol(sym));
    
    void set(<sf_symbol>field f)
    {
	if (simple_name(f.fld_symbol) == sn)
	    f.fld_symbol = Symbol(sym);
    }

    switch parent.stype {
    case Struct(x):
	List::iter(set, x);

    case Opt_struct(x):
	List::iter(set, x);

    case _:
	raise Invalid_argument("Module::reference_self_from_parent_struct");
    }
}

void reference_self_from_parent_union(var_symbol sym, ty_symbol parent)
{
    id sn = simple_name(Symbol(sym));
    
    void set(<var_symbol>field f)
    {
	if (simple_name(f.fld_symbol) == sn)
	    f.fld_symbol = Symbol(sym);
    }

    switch parent.stype {
    case Union(x):
	List::iter(set, x);

    case _:
	raise Invalid_argument("Module::reference_self_from_parent_union");
    }
}

void unserialize_var(module m, Xml::node n)
{
    let var_data = {fnc_level = -1, 
		    scope_level = -1, 
		    extern_name = Xml::prop_value("extern-name", n),
		    properties = 0,
		    parent_union = None} in {
	var_symbol sym = common_unserialize(m, n, None, var_data);
	Hashtbl::add(m.mod_vars, sym.name, sym);
	unxmlize_props(m, sym, Xml::find_element("props", n));
	
	switch sym.a.parent_union {
	case Some(pu):
	    reference_self_from_parent_union(sym, pu);
	case None:
	    skip;
	}
    }
}

void unserialize_struct_field(module m, Xml::node n)
{
    let tydef = lookup_tysym_q(m, Xml::prop_value("tydef", n)) in
    let sym = common_unserialize(m, n, None, ({sf_tydef = tydef})) in {
	Hashtbl::add(m.mod_struct_fields, sym.name, sym);
	reference_self_from_parent_struct(sym, tydef);
    }
}

void unserialize_depends_on(module m, Xml::node n)
{
    switch n {
    case Xml::Element("module", [], [("name", x)]):
	m.mod_depends_on = Cons(x, m.mod_depends_on);
    case _:
	raise Invalid_argument("Module::unserialize_depends_on");
    }
}

module create_module(id name)
{
    return {mod_name = name,
	    mod_vars = Hashtbl::create(13),
	    mod_tydefs = Hashtbl::create(13),
	    mod_struct_fields = Hashtbl::create(13),
	    mod_under_construction = true,
	    mod_depends_on = Nil};
}

module parse_module_interface(id name, string filename)
{
    Xml::node root = Xml::root_node(Xml::parse_file(filename, true));
    
    switch root {
    case Xml::Element("module", _, _):
	skip;
    
    case _:
	raise Not_found;
    }
    
    if (Xml::prop_value("name", root) != name) {
	error("file `" + filename + 
	      "' contains module interface for `" +
	      Xml::prop_value("name", root) +
	      "' instead of expected `" + name + "'");
	raise Choke;
    }
    
    module m = create_module(name);

    Hashtbl::add(open_modules, name, m);

    unserialize_all("tydefs", root, m, unserialize_type1);
    unserialize_all("tydefs", root, m, unserialize_type2);
    unserialize_all("vars", root, m, unserialize_var);
    unserialize_all("struct-fields", root, m, unserialize_struct_field);
    unserialize_all("depends-on", root, m, unserialize_depends_on);

    m.mod_under_construction = false;

    return m;
}

void check_name(id name)
{
    _ ok = true;
    
    def i(c) (int_of_char(c));
    
    def check_char(c)
    {
	if (!((i(c) >= i('a') && i(c) <= i('z')) ||
	      (i(c) >= i('A') && i(c) <= i('Z')) ||
	      (i(c) >= i('0') && i(c) <= i('9')) ||
	      (i(c) == i('_'))))
	    ok = false;
    }

    String::iter(check_char, name);

    if (!ok) {
	error("name `" + name + "' contains invalid characters");
	raise Choke;
    }
}

module open_module(id name)
{
    string filename = "";

    check_name(name);

    try {
	module m = Hashtbl::find(open_modules, name);
	if (m.mod_under_construction) {
	    error("module `" + name +
		  "' is mutally recursive with another module");
	    raise Choke;
	} else
	    return m;
    } with {
    case Not_found:
	skip;
    }

    if (empty_module != "" && name == empty_module) {
	_ m = create_module(name);
	m.mod_under_construction = false;
	Hashtbl::add(open_modules, name, m);
	return m;
    }
    
    try {
	let nm = String::uncapitalize(name) + ".xgi" in
	filename = find_in_path(search_path, nm);
    } with {
    case Not_found:
	error("cannot find compiled module interface for `" + name + "'");
	raise Choke;
    }

    try {
	return parse_module_interface(name, filename);	
    } with {
    case Xml::Parse_error(_):
	error("xml for module interface `" + name + "' in file `" +
	      filename + "' is not well formed");
	raise Choke;

    case Not_found:
	error("xml for module interface `" + name + "' in file `" +
	      filename + "' has bad structure");
	raise Choke;
    }
}


// Serialization stuff

<*(string, string)>list common_sym_props(<'a>symbol sym)
{
    return [("name", sym.name), 
	    ("file", sym.sloc.file),
	    ("line", itoa(sym.sloc.first_line)),
	    ("column", itoa(sym.sloc.first_column))];
}

void dump_node(Xml::node n)
{
    switch n {
    case Xml::Element(name, children, attrs):
	string soa(*(string, string) a)
	{
	    let (nn, v) = a in
	    return nn + "=\"" + v + "\"";
	}
	print_string("<" + name + " " + 
		     String::concat(" ", List::map(soa, attrs)) + ">\n");
	List::iter(dump_node, children);
	print_string("</"+name+">\n");
    case Xml::Text(s):
	print_string(s);
    }
}

Xml::node serialize_hash(<'a>Hashtbl::t h, string name, *(Xml::node ('a)) f)
{
    <Xml::node>list syms = Nil;
    
    void serialize(string, 'a sym)
    {
	syms = Cons(f(sym), syms);
    }

    Hashtbl::iter(h, serialize);

    let n = Xml::Element(name, syms, Nil) in {
//	dump_node(n);
	return n;
    }
}

Xml::node xmlize_props(var_symbol sym)
{
    _ props = [];

    _ add(_ n)
    {
	props = Cons(Xml::Element(n, [], []), props);
    }
   
    // any change here should be reflected in unxmlize_props() above
    if (has_property(sym, Is_union_field))
	add("is-union-field");
    if (has_property(sym, Is_lvalue))
	add("is-lvalue");
    if (has_property(sym, Is_initialized))
	add("is-initialized");
    if (has_property(sym, Is_global_var))
	add("is-global-var");
    if (has_property(sym, Is_exception))
	add("is-exception");
    if (has_property(sym, Is_exported))
	add("is-exported");
    if (has_property(sym, Is_extern_c))
	add("is-extern-c");
	
    switch sym.a.parent_union {
    case Some(pu):
	let e = Xml::Element("parent-union", [], [("value", pu.name)]) in
	props = Cons(e, props);

    case None:
	skip;
    }

    return Xml::Element("props", props, []);
}

Xml::node serialize_var(var_symbol sym)
{
    let props = Cons(("extern-name", sym.a.extern_name), common_sym_props(sym)),
	children = [xmlize_ty(sym.stype), xmlize_props(sym)] in
    return Xml::Element("var", children, props);
}

Xml::node serialize_tydef(ty_symbol sym)
{
    switch xmlize_ty_symbol(sym) {
    case Xml::Element("tydef", children, attrs):
	bool no_name_p(*(string, string) p)
	{
	    let (n, _) = p in
	    return n != "name";
	}
	let no_name = List::filter(no_name_p, attrs) in
	let nattrs = List::concat(no_name, common_sym_props(sym)) in
	return Xml::Element("tydef", children, nattrs);
    case _:
	return ice("Module::serialize_tydef(): " +
	       "xmlize_ty_symbol went crazy");
    }
}

Xml::node serialize_struct_field(sf_symbol sym)
{
    let props = Cons(("tydef", symbol_name(sym.a.sf_tydef)), 
		   common_sym_props(sym)),
	children = [xmlize_ty(sym.stype)] in
    return Xml::Element("struct-field", children, props);
}

Xml::node serialize_module(module m)
{
    Xml::node named_module(id n)
    {
	return Xml::Element("module", Nil, [("name", n)]);
    }
    
    let vars = serialize_hash(m.mod_vars, "vars", serialize_var),
	sfs = serialize_hash(m.mod_struct_fields, "struct-fields", 
			     serialize_struct_field),
	tydefs = serialize_hash(m.mod_tydefs, "tydefs", serialize_tydef),
	depends_on = Xml::Element("depends-on",
				  List::map(named_module, m.mod_depends_on),
				  Nil) in
    let n = Xml::Element("module", 
			 [vars, sfs, tydefs, depends_on],
			 [("name", m.mod_name)]) in
    {
	return n;
    }
}

// exported interface:
void save_module(module m, string filename)
{
    check_name(m.mod_name);
    Xml::node n = serialize_module(m);
    
    //dump_node(n);

    Stdio::FILE f = Stdio::fopen(filename, "w");
    
    try {
	Xml::save_to_stream_indented(Xml::make_doc(n), f);
    } finally {
	Stdio::fclose(f);
    }
}

// vim:sts=4:sw=4
